"""
This is using the Pollinations platform with contains an
AI image generator API and open-source model.

Source: https://pollinations.ai/
"""
from yta_general_utils.downloader import Downloader
from yta_general_utils.url import encode_url_parameter
from yta_general_utils.programming.parameter_validator import PythonValidator


def generate_image_with_pollinations(prompt: str, output_filename: str):
    """
    Generate an image with the Pollinations AI image generation model
    using the provided 'prompt' and stores it locally as 
    'output_filename'.
    """
    if not PythonValidator.is_string(prompt):
        raise Exception('Provided "prompt" parameter is not a valid prompt.')
    
    if not PythonValidator.is_string(output_filename):
        raise Exception('Provided "output_filename" parameter is not a valid output.')
    
    prompt = encode_url_parameter(prompt)

    # TODO: Make some of these customizable
    WIDTH = 1920
    HEIGHT = 1080
    # TODO: This seed should be a random value or
    # I will receive the same image with the same
    # prompt
    SEED = 43
    MODEL = 'flux'

    url = f'https://pollinations.ai/p/{prompt}?width={WIDTH}&height={HEIGHT}&seed={SEED}&model={MODEL}'

    Downloader.download_image(url, output_filename)

    return output_filename

"""
Check because there is also a model available for
download and to work with it (as they say here
https://pollinations.ai/):

    # Using the pollinations pypi package
    ## pip install pollinations

    import pollinations as ai

    model_obj = ai.Model()

    image = model_obj.generate(
        prompt=f'Awesome and hyperrealistic photography of a vietnamese woman... {ai.realistic}',
        model=ai.flux,
        width=1038,
        height=845,
        seed=43
    )
    image.save('image-output.jpg')

    print(image.url)
"""


def midpt(*args):
    xs,ys = 0,0
    for p in args:
        xs += p[0]
        ys += p[1]
    return xs/len(args),ys/len(args)

def distsum(*args):
    return sum([ ((args[i][0]-args[i-1][0])**2 + (args[i][1]-args[i-1][1])**2)**0.5 for i in range(1,len(args))])


from yta_general_utils.math.rate_functions import RateFunction
from math import pi

import numpy as np


# TODO: This method is duplicated in
# yta_multimedia\video\edition\effect\moviepy\t_function\__init__.py
def rate_function(t: float, duration: float, rate_func: type, *args, **kwargs):
    """
    You need to provide one of the functions of RateFunction class
    as the 'rate_func' parameter to be able to make it work, and 
    pass the needed args to it.
    """
    return rate_func(t / duration, *args, **kwargs)

def circles(t: float, duration: float, position: tuple, radius: int = 200, time_per_circle: float = 1, rate_func: type = RateFunction.linear):
        """
        Returns the (x, y) position tuple for the moviepy '.with_position()' effect,
        for each 't' provided, that will make the element move in circles with the
        provided 'radius'. The 'radius' parameter is the distance between the origin
        and the path the clip will follow. The 'cicle_time' is the time (in seconds)
        needed for a complete circle to be completed by the movement.

        If you provide the video duration as 'cicle_time', the video will make only
        one whole circle
        """
        circle_factor = rate_function(t % time_per_circle, time_per_circle, rate_func)

        # TODO: Do checkings
        return position[0] + radius * np.cos((circle_factor) * 2 * pi), position[1] + radius * np.sin((circle_factor) * 2 * pi)
        # Working code was this below:
        # return x + radius * np.cos((t / time_per_circle) * 2 * math.pi), y + radius * np.sin((t / time_per_circle) * 2 * math.pi)


from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_multimedia.utils.resize import get_cropping_points_to_keep_aspect_ratio
from yta_general_utils.programming.enum import YTAEnum as Enum
from moviepy.Clip import Clip
from moviepy import CompositeVideoClip
from moviepy.video.fx import Crop


class ResizeMode(Enum):
    """
    Enum class to encapsulate the different strategies to
    enlarge a video.
    """
    RESIZE_KEEPING_ASPECT_RATIO = 'resize_keeping_aspect_ratio'
    """
    This video is resized to fit the expected larger size
    by keeping the aspect ratio. Keeping the aspect ratio
    means that a part of the video can be lost because of
    cropping it.
    """
    RESIZE = 'resize'
    """
    The video is resized to fit the expected larger size.
    This won't keep the aspect ratio, so the whole video
    will be visible but maybe not properly. Use another
    option if possible.
    """
    FIT_LIMITING_DIMENSION = 'fit_limiting_dimension'
    """
    The video is resized to fit the most limiting dimension
    and is placed over a black background of the expected
    size.
    """
    BACKGROUND = 'background'
    """
    The video is just placed in the center, over a black
    background that has the desired dimensions. This will
    work exactly as FIT_ONE_DIMENSION if the video provided
    is larger than the expected size.
    """

def resize_video(video: Clip, size: tuple, resize_mode: ResizeMode = ResizeMode.RESIZE_KEEPING_ASPECT_RATIO):
    """
    Make the provided 'video' fit the also provided 'size' by 
    applying the given 'enlarge_mode' or 'enshort_mode' depending
    of the need.
    """
    # TODO: This generates a cyclic import issue
    #video = VideoParser.to_moviepy(video)

    resize_mode = ResizeMode.to_enum(resize_mode) if resize_mode is not None else ResizeMode.default()

    if (video.w, video.h) == size:
        return video

    # TODO: Maybe I could use a simulated image of provided
    # size and use the existing resizing methods
    if resize_mode == ResizeMode.RESIZE:
        video = video.resized(size)
    elif resize_mode == ResizeMode.RESIZE_KEEPING_ASPECT_RATIO:
        # We need to resize it first until we reach the greatest
        # dimension (width or height depending on the source element)
        original_ratio = video.w / video.h
        new_ratio = size[0] / size[1]

        if original_ratio > new_ratio:
            # Original video is wider than the expected one
            video = video.resized(height = size[1])
        elif original_ratio < new_ratio:
            # Original video is higher than the expected one
            video = video.resized(width = size[0])
        else:
            video = video.resized(size)

        # Now, with the new video resized, we look for the
        # cropping points we need to apply and we crop it
        top_left, bottom_right = get_cropping_points_to_keep_aspect_ratio((video.w, video.h), size)
        # Crop the video to fit the desired aspect ratio
        # TODO: Maybe avoid this if nothing to crop
        video = video.with_effects([Crop(width = bottom_right[0] - top_left[0], height = bottom_right[1] - top_left[1], x_center = video.w / 2, y_center = video.h / 2)])
        # Resize it to fit the desired 'size'
        video = video.resized(size)
    elif resize_mode == ResizeMode.FIT_LIMITING_DIMENSION:
        # We need to resize setting the most limiting dimension
        # to the one provided in the expected 'size' and then
        # place it over a background
        original_ratio = video.w / video.h
        new_ratio = size[0] / size[1]

        if original_ratio > new_ratio:
            # Original video is wider than the expected one
            video = video.resized(height = size[1])
        elif original_ratio < new_ratio:
            # Original video is higher than the expected one
            video = video.resized(width = size[0])
        else:
            video = video.resized(size)

        video = CompositeVideoClip([
            ClipGenerator.get_default_background_video(size, video.duration, False),
            video.with_position(('center', 'center'))
        ])
    elif resize_mode == ResizeMode.BACKGROUND:
        if video.w > size[0] or video.h > size[1]:
            video = resize_video(video, size, ResizeMode.FIT_LIMITING_DIMENSION)
        else:
            # Just place it in the center of a black background
            video = CompositeVideoClip([
                ClipGenerator.get_default_background_video(size, video.duration, False),
                video.with_position(('center', 'center'))
            ])

    return video


from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.duration import set_video_duration, ExtendVideoMode, EnshortVideoMode
from yta_multimedia.image.region.image_region import ImageRegion
from yta_general_utils.programming.parameter_validator import NumberValidator
from yta_general_utils.image.parser import ImageParser
from moviepy import ImageClip, CompositeVideoClip, VideoClip


# TODO: It would be interesting that the ImageGreenscreen,
# VideoGreenscreen, ImageAlphascreen and VideoAlphascreen
# inherit from this one instead of having an instance and
# using it, but I decided to start like this because of
# the inheritance problems I've experienced in the past 
# that delayed me too much.
class MaskedClipCreator:
    """
    Class created to encapsulate the way we handle regions with
    transparency in moviepy clips and to add other multimedia
    elements behind those regions to be seen through a new 
    CompositedVideoClip.

    This class is intended to be used by the Alphascreen and
    Greenscreen classes that are generating this kind of videos
    with alpha or green regions on the source files.
    """
    regions: list[ImageRegion] = []
    """
    The regions existing in the masked_clip
    """
    masked_clip: VideoClip = None
    """
    The masked clip we want to use to put some multimedia elements
    behind it with the transparency applied to let them be seen
    through the existing regions.
    """

    def __init__(self, regions, masked_clip):
        self.regions = regions
        self.masked_clip = masked_clip

    def from_image_to_image(self, image, output_filename: str = None):
        """
        This method returns a numpy representation of the image
        built by inserting the provided 'image' in this alphascreen.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        # TODO: This is not returning RGBA only RGB. You can use
        # the 'rgba' method from VideoFrameExtractor
        return self.from_images_to_image([image], output_filename)

    def from_images_to_image(self, images, output_filename: str = None):
        """
        This method returns a numpy representation of the image
        built by inserting the provided 'images' in this
        alphascreen.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        video = self.from_images_to_video(images, duration = 1 / 60)

        # TODO: Use VideoFrameExtractor
        if output_filename:
            video.save_frame(output_filename, t = 0)

        # TODO: This is not returning RGBA only RGB. You can use
        # the 'rgba' method from VideoFrameExtractor
        return video.get_frame(t = 0)

    def from_image_to_video(self, image, duration: float, output_filename: str = None):
        """
        This method returns a CompositeVideoClip with the provided
        'image' fitting the first alphascreen area and centered on
        those areas by applying a mask that let them be seen
        through that mask.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        return self.from_images_to_video([image], duration, output_filename)
    
    def from_images_to_video(self, images, duration: float, output_filename: str = None):
        """
        This method returns a CompositeVideoClip with the provided
        'images' fitting the different alphascreen areas and
        centered on those areas by applying a mask that let them be
        seen through that mask.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        self._validate_enough_elements_for_regions(images)

        if not NumberValidator.is_positive_number(duration):
            raise Exception('No valid "duration" parameter provided. It must be a positive number.')

        for image in images:
            image = ImageParser.to_numpy(image)
        
        # TODO: Use the 'transparent' parameter (?)
        videos = [ImageClip(image, duration = duration).with_fps(60) for image in images]

        return self.from_videos_to_video(videos, output_filename)
    
    def from_video_to_video(self, video, output_filename: str = None):
        """
        This method returns a CompositeVideoClip with the provided
        'video' fitting in the alphascreen area and centered on it
        by applying a mask that let it be seen through that mask.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        return self.from_videos_to_video([video], output_filename)

    def from_videos_to_video(self, videos, output_filename: str = None):
        """
        This method is pretend to be used by the Alphascreen or
        Greenscreen classes to simplify the process and code.

        This method returns a CompositeVideoClip with the provided
        'videos' fitting the different alphascreen areas and
        centered on those areas by applying a mask that let them be
        seen through that mask.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        self._validate_enough_elements_for_regions(videos)

        longest_duration = 0
        for video in videos:
            video = VideoParser.to_moviepy(video)
            if video.duration > longest_duration:
                longest_duration = video.duration

        for index, video in enumerate(videos):
            videos[index] = self.regions[index].place_video_inside(video)

        # TODO: What if video is not as longer as 'longest_duration' (?)
        masked_clip = self.masked_clip.with_duration(longest_duration)

        return self._build_composite_clip(videos, masked_clip, output_filename)


    def _validate_enough_elements_for_regions(self, elements):
        """
        Raises an exception if the provided amount of 'elements' is 
        greater or less than the amount of alpha regions.
        """
        if len(elements) > len(self.regions) or len(elements) < len(self.regions):
            raise Exception(f'There are more or less elements provided ({str(len(elements))}) than available masked regions ({str(len(self.regions))}).')

    def _build_composite_clip(self, videos, masked_clip, output_filename: str = None):
        """
        Builds the CompositeVideoClip that includes the provided 'videos'
        and the also provided 'alpha_clip' to build the desired video with
        alpha regions filled with the videos.
        """
        # TODO: Please private method
        # As this is for internal use I consider that 'videos' and
        # 'alpha_clip' are valid ones and ready to be used at this point

        # TODO: Provided videos can be shorther than the alphascreen
        # or the alphascreen can be shorter than the videos, so we
        # need an strategy to follow. By now I'm forcing all the 
        # videos to fit the alphascreen duration by shortening or
        # enlarging them.
        for index, _ in enumerate(videos):
            videos[index] = set_video_duration(videos[index], masked_clip.duration, ExtendVideoMode.FREEZE_LAST_FRAME, EnshortVideoMode.CROP)

        composite_clip = CompositeVideoClip([
            *videos,
            masked_clip
        ], size = masked_clip.size)

        if not composite_clip.fps:
            composite_clip = composite_clip.with_fps(60)

        if output_filename:
            composite_clip.write_videofile(output_filename)

        return composite_clip



from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.image.parser import ImageParser
from typing import Union

import numpy as np


class FrameMaskingMethod(Enum):
    MEAN = 'mean'
    """
    Calculate the mean value of the RGB pixel color
    values and uses it as a normalized value between
    0.0 and 1.0 to set as transparency.
    """
    PURE_BLACK_AND_WHITE = 'pure_black_and_white'
    """
    Apply a threshold and turn pixels into pure black
    and white pixels, setting them to pure 1.0 or 0.0
    values to be completely transparent or opaque.
    """
    # TODO: Add more methods
    # TODO: Create a method that applies a 'remove_background'
    # and then the remaining shape as a mask that can be zoomed
    # so we obtain a nice effect

    def to_mask_frame(self, frame: np.ndarray):
        """
        Process the provided video normal 'frame' according to this
        type of masking processing method and turns it into a frame
        that can be used as a mask frame.
        """
        frame = ImageParser.to_numpy(frame)

        if not VideoFrameValidator.frame_is_video_frame(frame):
            raise Exception('The provided "frame" is not actually a moviepy normal video frame.')

        if self == FrameMaskingMethod.MEAN:
            frame = np.mean(frame, axis = -1) / 255.0
        elif self == FrameMaskingMethod.PURE_BLACK_AND_WHITE:
            frame = pure_black_and_white_image_to_moviepy_mask_numpy_array(frame_to_pure_black_and_white_image(frame))

        return frame

class NumpyFrameHelper:
    """
    Class to encapsulate functionality related to numpy
    frames. Numpy frames are frames with width and height,
    and 1 or 3 values per pixel (per cell).
    """

    @staticmethod
    def normalize(frame: np.ndarray, do_check: bool = True):
        """
        Normalize the frame if not normalized.
        """
        if do_check and NumpyFrameHelper.is_rgb_not_normalized(frame) or NumpyFrameHelper.is_alpha_not_normalized(frame) or NumpyFrameHelper.is_rgba_not_normalized(frame):
            frame = frame / 255.0

        return frame

    @staticmethod
    def denormalize(frame: np.ndarray, do_check: bool = True):
        """
        Denormalize the frame if normalized.
        """
        if do_check and NumpyFrameHelper.is_rgb_normalized(frame) or NumpyFrameHelper.is_alpha_normalized(frame) or NumpyFrameHelper.is_rgba_normalized(frame):
            frame = (frame * 255).astype(np.uint8)

        return frame
    
    @staticmethod
    def is_normalized(frame: np.ndarray):
        """
        Check if the provided frame is a a normalized one, which
        means that its type is .float64 or .float32 and that all
        values are between 0.0 and 1.0.
        """
        frame = ImageParser.to_numpy(frame)

        return frame.dtype in (np.float64, np.float32) and np.all((frame >= 0.0) & (frame <= 1.0))

    @staticmethod
    def is_not_normalized(frame: np.ndarray):
        """
        Check if the provided frame is not a normalized one, which
        means that its type is .uint8 and that all values are 
        between 0 and 255.
        """
        frame = ImageParser.to_numpy(frame)
        
        return frame.dtype == np.uint8 and np.all((frame >= 0) & (frame <= 255))

    @staticmethod
    def is_rgb(frame: np.ndarray, is_normalized: Union[None, bool] = None):
        """
        Check if the provided 'frame' is an RGB frame, which means
        that its dimension is 3 and its shape is also 3 per pixel.

        If 'is_normalized' is provided, it will check if the frame
        is normalized or not according to the boolean value passed
        as parameter.
        """
        frame = ImageParser.to_numpy(frame)

        is_rgb = frame.ndim == 3 and frame.shape[2] == 3

        if is_normalized is None:
            return is_rgb
        
        if is_normalized:
            return is_rgb and NumpyFrameHelper.is_normalized(frame)
        
        return is_rgb and NumpyFrameHelper.is_not_normalized(frame)

    @staticmethod
    def is_rgba(frame: np.ndarray, is_normalized: Union[None, bool] = None):
        """
        Check if the provided 'frame' is an RGBA frame, which means
        that its dimension is 3 and its shape is 4 per pixel.

        If 'is_normalized' is provided, it will check if the frame
        is normalized or not according to the boolean value passed
        as parameter.

        TODO: This is not actually a frame we can use in moviepy
        videos, but it could be a frame we build to later decompose
        in clip and mask clip, so I keep the code. Maybe it is 
        useless in the future and thats why this is a TODO.
        """
        frame = ImageParser.to_numpy(frame)

        is_rgba = frame.ndim == 3 and frame.shape[2] == 4

        if is_normalized is None:
            return is_rgba
        
        if is_normalized:
            return is_rgba and NumpyFrameHelper.is_normalized(frame)
        
        return is_rgba and NumpyFrameHelper.is_not_normalized(frame)
    
    @staticmethod
    def is_alpha(frame: np.ndarray, is_normalized: Union[None, bool] = None):
        """
        Check if the provided 'frame' is an alpha frame, which means
        that its dimension is 2 because there is only one single
        value per pixel.

        
        If 'is_normalized' is provided, it will check if the frame
        is normalized or not according to the boolean value passed
        as parameter.
        """
        frame = ImageParser.to_numpy(frame)

        # if not PythonValidator.is_numpy_array(frame):
        #     raise Exception('The provided "frame" parameter is not a numpy array.')

        is_alpha = frame.ndim == 2

        if is_normalized is None:
            return is_alpha
        
        if is_normalized:
            return is_alpha and NumpyFrameHelper.is_normalized(frame)
        
        return is_alpha and NumpyFrameHelper.is_not_normalized(frame)

    @staticmethod
    def is_rgb_not_normalized(frame: np.ndarray):
        """
        Check if the provided 'frame' is a numpy array of
        ndim = 3, dtype = np.uint8 and all the values (3)
        are between 0 and 255.
        """
        return NumpyFrameHelper.is_rgb(frame, is_normalized = False)
    
    @staticmethod
    def is_rgb_normalized(frame: np.ndarray):
        """
        Check if the provided 'frame' is a numpy array of
        ndim = 3, dtype = np.float64|np.float32 and all 
        the values (3) are between 0.0 and 1.0.
        """
        return NumpyFrameHelper.is_rgb(frame, is_normalized = True)

    @staticmethod
    def is_rgba_normalized(frame: np.ndarray):
        return NumpyFrameHelper.is_rgba(frame, is_normalized = True)
    
    @staticmethod
    def is_rgba_not_normalized(frame: np.ndarray):
        return NumpyFrameHelper.is_rgba(frame, is_normalized = False)

    @staticmethod
    def is_alpha_normalized(frame: np.ndarray):
        return NumpyFrameHelper.is_alpha(frame, is_normalized = True)

    @staticmethod
    def is_alpha_not_normalized(frame: np.ndarray):
        return NumpyFrameHelper.is_alpha(frame, is_normalized = False)

    @staticmethod
    def as_rgb(frame: np.ndarray, do_normalize: bool = False):
        """
        Turn the provided 'frame' to a normal (rgb) frame,
        normalized or not according to the provided as
        'do_normalize' parameter.

        This method will return a numpy array containing 3
        values for each pixel, and each one for them will be
        from 0.0 to 1.0 if normalized, or from 0 to 255 if
        not normalized.

        A default moviepy frame is a numpy array of 3 values
        per pixel from 0 to 255.
        """
        if NumpyFrameHelper.is_alpha_normalized(frame):
            frame = np.stack((frame, frame, frame), axis = -1)
            if not do_normalize:
                frame = NumpyFrameHelper.denormalize(frame, do_check = False)
        if NumpyFrameHelper.is_alpha_not_normalized(frame):
            frame = np.stack((frame, frame, frame), axis = -1)
            if do_normalize:
                frame = NumpyFrameHelper.normalize(frame, do_check = False)
        elif NumpyFrameHelper.is_rgb_normalized(frame):
            if not do_normalize:
                frame = NumpyFrameHelper.denormalize(frame, do_check = False)
        elif NumpyFrameHelper.is_rgb_not_normalized(frame):
            if do_normalize:
                frame = NumpyFrameHelper.normalize(frame, do_check = False)
        elif NumpyFrameHelper.is_rgba_normalized(frame):
            frame = frame[:, :, :3]
            if not do_normalize:
                frame = NumpyFrameHelper.denormalize(frame, do_check = False)
        elif NumpyFrameHelper.is_rgba_not_normalized(frame):
            frame = frame[:, :, :3]
            if do_normalize:
                frame = NumpyFrameHelper.normalize(frame, do_check = False)
        else:
            raise Exception('The provided "frame" is not recognized as a valid frame (RGB, RGBA or alpha).')

        return frame
    
    @staticmethod
    def as_alpha(frame: np.ndarray, do_normalize: bool = True, masking_method: FrameMaskingMethod = FrameMaskingMethod.MEAN):
        """
        Turn the provided 'frame' to an alpha frame, normalized
        or not according to the provided as 'do_normalize'
        parameter.

        This method will return a numpy array containing one
        single value for each pixel, that will be from 0.0 to
        1.0 if normalized, or from 0 to 255 if not normalized.

        A default moviepy mask frame is a numpy array of one
        single value per pixel from 0.0 to 1.0.

        The 'masking_method' will determine the method that is
        needed to be used to turn the normal frame into a mask
        frame.
        """
        masking_method = FrameMaskingMethod.to_enum(masking_method)

        if NumpyFrameHelper.is_alpha_normalized(frame):
            if not do_normalize:
                frame = NumpyFrameHelper.denormalize(frame, do_check = False)
        if NumpyFrameHelper.is_alpha_not_normalized(frame):
            frame = np.stack((frame, frame, frame), axis = -1)
            if do_normalize:
                frame = NumpyFrameHelper.normalize(frame, do_check = False)
        elif NumpyFrameHelper.is_rgb_normalized(frame):
            frame = masking_method.to_mask_frame(frame)
            if not do_normalize:
                frame = NumpyFrameHelper.denormalize(frame, do_check = False)
        elif NumpyFrameHelper.is_rgb_not_normalized(frame):
            frame = masking_method.to_mask_frame(frame)
            if do_normalize:
                frame = NumpyFrameHelper.normalize(frame, do_check = False)
        elif NumpyFrameHelper.is_rgba_normalized(frame):
            frame = frame[:, :, :3]
            frame = masking_method.to_mask_frame(frame)
            if not do_normalize:
                frame = NumpyFrameHelper.denormalize(frame, do_check = False)
        elif NumpyFrameHelper.is_rgba_not_normalized(frame):
            frame = frame[:, :, :3]
            frame = masking_method.to_mask_frame(frame)
            if do_normalize:
                frame = NumpyFrameHelper.normalize(frame, do_check = False)
        else:
            raise Exception('The provided "frame" is not recognized as a valid frame (RGB, RGBA or alpha).')

        return frame

    def invert(frame: np.ndarray):
        """
        Invert the provided array according to if it is a normalized
        or a not normalized one.
        """
        if NumpyFrameHelper.is_normalized():
            frame = 1.0 - frame
        elif NumpyFrameHelper.is_not_normalized():
            frame = 255 - frame
        else:
            raise Exception('The provided "frame" is not a normalized array nor a not normalized one.')
        
        return frame



WHITE = [255, 255, 255]
BLACK = [0, 0, 0]

def is_pure_black_and_white_image(image):
    """
    Check if the provided 'image' only contains pure 
    black ([0, 0, 0]) and white ([255, 255, 255]) colors.
    """
    image = ImageParser.to_numpy(image)

    if np.any(~np.all((image == WHITE) | (image == BLACK), axis = -1)):
        return False
    
    return True

# TODO: Should I combine these 2 methods below in only 1 (?)
def pure_black_and_white_image_to_moviepy_mask_numpy_array(image):
    """
    Turn the received 'image' (that must be a pure black
    and white image) to a numpy array that can be used as
    a moviepy mask (by using ImageClip).

    This is useful for static processed images that we 
    want to use as masks, such as frames to decorate our
    videos.
    """
    image = ImageParser.to_numpy(image)

    if not is_pure_black_and_white_image(image):
        raise Exception(f'The provided "image" parameter "{str(image)}" is not a black and white image.')

    # Image to a numpy parseable as moviepy mask
    mask = np.zeros(image.shape[:2], dtype = int) # 3col to 1col
    mask[np.all(image == WHITE, axis = -1)] = 1 # white to 1 value

    return mask

def frame_to_pure_black_and_white_image(frame: np.ndarray):
    """
    Process the provided moviepy clip mask frame (that
    must have values between 0.0 and 1.0) or normal clip
    frame (that must have values between 0 and 255) and
    convert it into a pure black and white image (an
    image that contains those 2 colors only).

    This method returns a not normalized numpy array of only
    2 colors (pure white [255, 255, 255] and pure black
    [0, 0, 0]), perfect to turn into a mask for moviepy clips.

    This is useful when handling an alpha transition video 
    that can include (or not) an alpha layer but it is also
    clearly black and white so you transform it into a mask
    to be applied on a video clip.
    """
    frame = ImageParser.to_numpy(frame)

    if not VideoFrameValidator.frame_is_moviepy_frame(frame):
        raise Exception('The provided "frame" parameter is not a moviepy mask clip frame nor a normal clip frame.')
    
    if VideoFrameValidator.frame_is_video_frame(frame):
        # TODO: Process it with some threshold to turn it
        # into pure black and white image (only those 2
        # colors) to be able to transform them into a mask.
        threshold = 220
        white_pixels = np.all(frame >= threshold, axis = -1)

        # Image to completely and pure black
        new_frame = np.array(frame)
        
        # White pixels to pure white
        new_frame[white_pixels] = WHITE
        new_frame[~white_pixels] = BLACK
    elif VideoFrameValidator.frame_is_mask_frame(frame):
        transparent_pixels = frame == 1

        new_frame = np.array(frame)
        
        # Transparent pixels to pure white
        new_frame[transparent_pixels] = WHITE
        new_frame[~transparent_pixels] = BLACK

    return new_frame

class VideoFrameValidator:
    """
    Class to simplify the video frame validation process.
    """
    @staticmethod
    def frame_is_moviepy_frame(frame: np.ndarray):
        """
        Check if the provided 'frame' (that is a numpy array) is
        recognize as a normal or mask frame of the moviepy library.
        """
        return VideoFrameValidator.frame_is_video_frame(frame) or VideoFrameValidator.frame_is_mask_frame(frame)

    @staticmethod
    def frame_is_video_frame(frame: np.ndarray):
        """
        Checks if the provided 'frame' numpy array is recognized as
        a frame of a normal moviepy video with values between 0 and
        255.

        This numpy array should represent a frame of a clip.
        
        A non-modified clip is '.ndim = 3' and '.dtype = np.uint8'.
        """
        return NumpyFrameHelper.is_rgb_not_normalized(frame)
        
    @staticmethod
    def frame_is_mask_frame(frame: np.ndarray):
        """
        Checks if the provided 'frame' numpy array is recognized as
        an original moviepy mask clip with values between 0 and 1.
        This numpy array should represent a frame of a mask clip.
        
        A non-modified mask clip is '.ndim = 2' and '.dtype = np.float64'.
        """
        return NumpyFrameHelper.is_alpha_normalized(frame)


# TODO: Maybe a NumpyHelper to make reusable methods
# such as 'as_zeros(array)' to get a similar one but
# filled with zeros, to calculate the mean


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.frames.video_frame_extractor import VideoFrameExtractor
from yta_multimedia.video.generation import generate_video_from_image
from yta_general_utils.programming.parameter_validator import NumberValidator
from moviepy.Clip import Clip


class FreezeLastFrameEffect(Effect):
    """
    Freeze the last frame of the provided 'video' and returns
    a new video of the provided 'duration' that is that frame
    freezed.
    """
    def apply(self, video: Clip, duration: float = 1) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlurEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        duration = duration if duration is not None else 1

        if not NumberValidator.is_positive_number(duration, False):
            raise Exception(f'The provided "duration" parameter "{str(duration)}" is not valid.')

        frame = VideoFrameExtractor.get_frame_by_time(video, video.duration)
        frame_freezed_video = generate_video_from_image(frame, duration)

        return frame_freezed_video

    def apply_over_video(self):
        pass


from yta_multimedia.audio.parser import AudioParser
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.file.filename import ensure_file_extension
from yta_general_utils.programming.enum import YTAEnum as Enum


class AudioExtension(Enum):
    """
    Enum class to encapsulate the accepted audio extensions for our
    system.
    """
    # TODO: Maybe interconnect with 'ffmpeg_handler.py' Enums
    MP3 = 'mp3'
    WAV = 'wav'
    M4A = 'm4a'
    WMA = 'wma'
    CD = 'cd'
    OGG = 'ogg'
    AIF = 'aif'
    # TODO: Check which extensions are valid for the AudioSegment
    # and the 'export' method to be able to classify AudioExtension
    # enums in AudioSegmentAudioExtension or similar because we
    # should also have AudioExtension for the FfmpegHandler...

class AudioConverter:
    """
    Class to simplify and encapsulate the functionality related to
    audio conversion.
    """
    @staticmethod
    def to(audio, extension: AudioExtension, output_filename: str):
        """
        This method converts the provided 'audio' to a audio with
        the provided 'extension' by storing it locally as the 
        provided 'output_filename' (or as a temporary file if not
        provided), and returns the new audio and the filename.

        This method returns two values: audio, filename
        """
        audio = AudioParser.to_audiosegment(audio)
        extension = AudioExtension.to_enum(audio)

        if not output_filename:
            # TODO: Replace this when not exporting needed
            output_filename = create_temp_filename(f'tmp_converted_sound.{extension.value}')
        else:
            output_filename = ensure_file_extension(output_filename, extension.value)

        audio.export(output_filename, format = extension.value)
        audio = AudioParser.to_audiosegment(output_filename)

        return audio, output_filename

    @staticmethod
    def to_wav(audio, output_filename: str):
        """
        This method converts the provided 'audio' to a wav audio
        by storing it locally as the provided 'output_filename'
        (or as a temporary file if not provided), and returns the
        new audio and the filename.

        This method returns two values: audio, filename
        """
        return AudioConverter.to(audio, AudioExtension.WAV, output_filename)
    
    @staticmethod
    def to_mp3(audio, output_filename: str):
        """
        This method converts the provided 'audio' to a mp3 audio
        by storing it locally as the provided 'output_filename'
        (or as a temporary file if not provided), and returns the
        new audio and the filename.

        This method returns two values: audio, filename
        """
        return AudioConverter.to(audio, AudioExtension.MP3, output_filename)


from yta_multimedia.video.generation.manim.utils.dimensions import ManimDimensions
from yta_multimedia.video.generation.manim.classes.loading_bar.parameter_classes.loading_bar_image import LoadingBarImage
from manim import *
from PIL import Image


class LoadingBarMobject(Mobject):
    def __init__(self, x: int = 0, y: int = -400, width: int = 1400, height: int = 30, unreached_color: ManimColor = GREEN_B, reached_color: ManimColor = GREEN_A):
        """
        The 'x', 'y' and 'width' parameters must be in pixels. Remember that
        we use the scene as a 1920x1080 pixels screen.
        """
        if x == None or x < -960 or x > 960:
            raise Exception('"x" must be an "int" number between [-960, 960]')
        
        if y == None or y < -540 or y > 540:
            raise Exception('"y" must be an "int" number between [-540, 540]')

        if width == None or width <= 0 or width > 1920:
            raise Exception('"width" must be an "int" number between [0, 1920]')
        
        if height == None or height <= 0 or height > 1080:
            raise Exception('"width" must be an "int" number between [0, 1080]')
        
        # TODO: Check colors and raise Exception

        # We transform pixels to manim values and set parameters
        self.initial_x = ManimDimensions.width_to_manim_width(x)
        self.initial_y = ManimDimensions.height_to_manim_height(y)
        self.initial_width = ManimDimensions.width_to_manim_width(width)
        self.initial_height = ManimDimensions.height_to_manim_height(height)
        self.unreached_color = unreached_color
        self.reached_color = reached_color
        self.loading_bar_unreached = None
        self.loading_bar_reached = None
        # mobjects
        self.loading_bar_unreached = None
        self.loading_bar_reached = None
        self.image_mobjects = []

        # Generate the basic loading bar elements (two rectangles)
        self.__reset_bar()

    def __reset_bar(self):
        """
        Moves the bar to the start position. This instantiates the objects
        and moves them to the start position (0%).
        """
        if not self.loading_bar_unreached:
            self.loading_bar_unreached = Rectangle(width = self.initial_width, height = self.initial_height)
            self.loading_bar_unreached.set_fill(self.unreached_color, opacity = 1).set_stroke(self.unreached_color, width = 0).move_to((self.initial_x, self.initial_y, 0))
            
        if not self.loading_bar_reached:
            self.loading_bar_reached = Rectangle(width = ManimDimensions.width_to_manim_width(1), height = self.initial_height)
            self.loading_bar_reached.set_fill(self.reached_color, opacity = 1).set_stroke(self.reached_color, width = 0).move_to((self.loading_bar_unreached.get_x() - (self.loading_bar_unreached.width / 2 + self.loading_bar_reached.width / 2), self.loading_bar_unreached.get_y(), 0))
        else:
            self.loading_bar_reached.stretch_to_fit_width(ManimDimensions.width_to_manim_width(1)).move_to((self.loading_bar_unreached.get_x() - (self.loading_bar_unreached.width / 2 + self.loading_bar_reached.width / 2), self.loading_bar_unreached.get_y(), 0))

    def __calculate_from_percentage_to_percentage(self, start_percentage: int, end_percentage: int):
        """
            Calculates the start width, the end width and the difference between
            both. It will return an object containing 'width_at_start',
            'width_at_end' y 'width_difference'.

            It also returns 'x_at_start' and 'x_at_end' that are the x positions
            for those percentages. These are useful for images.
        """
        width_at_start = start_percentage * 0.01 * self.initial_width
        x_at_start = -self.initial_width / 2 + width_at_start
        width_at_end = end_percentage * 0.01 * self.initial_width
        x_at_end = -self.initial_width / 2 + width_at_end
        width_difference = width_at_end - width_at_start

        if width_at_start == 0:
            # We need to use at least 1 pixel of size
            width_at_start = ManimDimensions.width_to_manim_width(1)

        return {
            'width_at_start': width_at_start,
            'x_at_start': x_at_start,       
            'width_at_end': width_at_end,
            'x_at_end': x_at_end,
            'width_difference': width_difference
        }

    def __move_bar_to_percentage(self, percentage: int):
        # We reset bar to set to the starting position
        self.__reset_bar()

        if percentage == 0:
            return

        calculations = self.__calculate_from_percentage_to_percentage(0, percentage)

        # Move the bar to the new start position according to percentages
        self.loading_bar_reached.stretch_to_fit_width(calculations['width_at_end']).shift(RIGHT * (calculations['width_difference'] / 2))

    def get_mobjects(self):
        """
        Returns all the existing mobjects in this mobject to be able to add
        them to the scene.
        """
        mobjects = [
            self.loading_bar_unreached,
            self.loading_bar_reached
        ]

        if self.image_mobjects and len(self.image_mobjects) > 0:
            mobjects += self.image_mobjects

        return mobjects

    def get_animation(self, images: list[LoadingBarImage] = [], start_percentage: int = 0, end_percentage: int = 100, duration: float = 5):
        """
        Returns the animation to be played in the scene with the 'self.play'
        method. This allows us to execute it outside and to mix this animation
        with others. It is important to provide the exact duration to be able
        to build the correct animation.
        """
        if start_percentage == None or start_percentage < 0 or start_percentage > 100:
            raise Exception('"start_percentage" parameter must be a valid int between [0, 100]')
        
        if end_percentage == None or end_percentage < 0 or end_percentage > 100:
            raise Exception('"start_percentage" parameter must be a valid int between [0, 100]')
        
        if duration == None or duration <= 0 or duration > 120:
            raise Exception('"duration" parameter must be a valid float between (0, 120]')
        
        # TODO: Check if 'images' is provided and valid
        if images == None:
            images = []
        
        # We move the bar to the start percentage position
        self.__move_bar_to_percentage(start_percentage)
        
        # Calculate from start to end
        calculations = self.__calculate_from_percentage_to_percentage(start_percentage, end_percentage)

        self.image_mobjects = []

        animations_group = []
        for image in images:
            # TODO: This is not good, please, handle this in a better way
            image_data = Image.open(image.image_filename).resize((64, 64))

            image_calculations = self.__calculate_from_percentage_to_percentage(image.start_percentage, image.end_percentage)

            image_mobject = ImageMobject(np.array(image_data)).set_z_index(0).move_to((image_calculations['x_at_start'], self.loading_bar_reached.get_y() + image.y, 0))
            self.image_mobjects.append(image_mobject)

            if image.start_percentage != image.end_percentage:
                # We make the animation to move the image to destination
                animations_group.append(image_mobject.animate.shift(RIGHT * (image_calculations['x_at_end'] - image_mobject.get_x())))
            else:
                # We just move static images to their positions without animation
                image_mobject.move_to((image_mobject.get_x(), image_mobject.get_y(), 0))

        # We make the main animation, the bar moving
        animations_group.append(self.loading_bar_reached.animate.stretch_to_fit_width(calculations['width_at_end']).shift(RIGHT * (calculations['width_difference'] / 2)))

        return animations_group


from yta_multimedia.video.generation.manim.constants import SCENE_HEIGHT
from yta_multimedia.video.generation.manim.utils.dimensions import ManimDimensions
from yta_multimedia.video.parser import VideoParser
from yta_general_utils.temp import create_temp_filename
from moviepy import AudioFileClip
from manim import *


"""
Docummentation here: 
    https://docs.manim.community/en/stable/

Useful links:
- https://www.youtube.com/watch?v=KHGoFDB-raE (+1h of using SVGs and drawing and animating)
- https://www.youtube.com/watch?v=bCsk6hnMO5w   (Mobjects and animations)
            -> https://github.com/mphammer/Manim-Mobjects-and-Animations/blob/main/animations.py
- https://www.youtube.com/watch?v=5qj3b7DY5oA   (Mobjects and animations)
            -> https://github.com/mphammer/Manim-Mobjects-and-Animations/blob/main/mobjects.py

Command to throw:
    manim PYTHON_FILE.py CLASS_NAME -pqm
"""

"""
Interesting:
    - https://docs.manim.community/en/stable/examples.html (some examples)
    - https://medium.com/@andresberejnoi/data-visualization-bar-chart-animations-with-manim-andres-berejnoi-75ece91a2da4 (bar graphs)
"""

# TODO: Is all this needed (?)
def manim_alpha_animation_to_videoclip(video):
    """
    Reads the provided 'video' manim generated .mov animation,
    makes the audio fix and returns a VideoFileClip object to
    work with.

    I don't know how o where this bug is done, but when using
    directly a alpha manim animation that has any sound, with
    the moviepy library, the final audio is different and
    doesn't fit the original timing.

    This method extracts the original audio, turns it into a
    new mp3 file and set that new audio file as the video 
    audio, apparently fixing the bug.
    """
    video = VideoParser.to_moviepy(video, do_include_mask = True)

    MP3_EXPORTED_FILENAME = create_temp_filename('manim_exported_audio.mp3')
    video.audio.write_audiofile(MP3_EXPORTED_FILENAME)

    try:
        # Extract the audio
        #AudioSegment.from_file(video_filename).export(MP3_EXPORTED_FILENAME, format = 'mp3')

        # Replace the original audio in moviepy video with this extracted one
        video.audio = AudioFileClip(MP3_EXPORTED_FILENAME)
    except:
        pass

    return video

# TODO: Maybe this one needs to be moved to a text handler
def fitting_text(text, width_to_fit: float = 1920, fill_opacity: float = 1, stroke_width: float = 0, color: ParsableManimColor = None, font_size: float = DEFAULT_FONT_SIZE, line_spacing: float = -1, font: str = '', slant: str = NORMAL, weight: str = NORMAL, t2c: dict[str, str] = None, t2f: dict[str, str] = None, t2g: dict[str, tuple] = None, t2s: dict[str, str] = None, t2w: dict[str, str] = None, gradient: tuple = None, tab_width: int = 4, warn_missing_font: bool = True, height: float = None, width: float = None, should_center: bool = True, disable_ligatures: bool = False, **kwargs):
    width_to_fit = ManimDimensions.width_to_manim_width(width_to_fit)

    txt_width_fitted = Text(text, fill_opacity, stroke_width, color, font_size, line_spacing, font, slant, weight, t2c, t2f, t2g, t2s, t2w, gradient, tab_width, warn_missing_font, height, width, should_center, disable_ligatures, **kwargs).scale_to_fit_width(width_to_fit)
    # I use a margin of 100 pixels so avoid being just in the borders
    txt_height_fitted = Text(text, fill_opacity, stroke_width, color, font_size, line_spacing, font, slant, weight, t2c, t2f, t2g, t2s, t2w, gradient, tab_width, warn_missing_font, height, width, should_center, disable_ligatures, **kwargs).scale_to_fit_height(SCENE_HEIGHT - ManimDimensions.height_to_manim_height(100))

    # As it is a 16:9 proportion, the height is the measure that limits the most
    if txt_height_fitted.font_size < txt_width_fitted.font_size:
        return txt_height_fitted
    return txt_width_fitted

def fullscreen_image(filename, scale_to_resolution: int = QUALITIES[DEFAULT_QUALITY]["pixel_height"], invert: bool = False, image_mode: str = 'RGBA', **kwargs):
    """
    Returns an ImageMobject that fits the provided 'width_to_fit' ignoring height. This is useful
    if you want an Image that fills the whole screen width.
    """
    image_width_fitted = ImageMobject(filename, scale_to_resolution, invert, image_mode, **kwargs).scale_to_fit_width(ManimDimensions.width_to_manim_width(1920))
    image_height_fitted = ImageMobject(filename, scale_to_resolution, invert, image_mode, **kwargs).scale_to_fit_height(ManimDimensions.height_to_manim_height(1080))

    # We want the image that occupies the whole screen
    if ManimDimensions.manim_height_to_height(image_width_fitted.height) >= 1080:
        return image_width_fitted
    
    return image_height_fitted

def preprocess_image(image: ImageMobject):
    """
    This method processes images bigger than our 1920x1080 dimensions and returns it
    scaled down to fit those dimensions. You should use this method as the first one
    when working with ImageMobjects, and then scaling it down as much as you need.
    """
    if ManimDimensions.manim_width_to_width(image.width) > 1920:
        image.scale_to_fit_width(ManimDimensions.width_to_manim_width(1920))
    if ManimDimensions.manim_height_to_height(image.height) > 1080:
        image.scale_to_fit_height(ManimDimensions.height_to_manim_height(1080))

    return image

# TODO: Remove this below as it is useless
# def add_manim_animation_to_video(original_video_filename, original_video_duration, manim_animation_name, manim_animation_parameters, output_filename):
#     """
#     Generates a manim animation video file with the provided 'manim_animation_name' and
#     'manim_animation_parameters', and puts it in the foreground of the original video
#     provided.

#     'original_video_duration' parameter is passed to crop the original video to that
#     duration in seconds if it is interesting for the final animation.

#     This is a method that must be used to add any manim animation to any app-generated
#     video.
#     """
#     from utils.video_utils import add_manin_animation_to_video as add_manim_animation_to_videoclip
#     TMP_MANIM_ANIMATION_FILENAME = PROJECT_ABSOLUTE_PATH + 'tmp_manim_animation.mov'

#     # TODO: Check that 'manim_animation_name' exist (is valid)

#     # We generate the animation video file
#     generate_manim_animation(manim_animation_name, manim_animation_parameters, TMP_MANIM_ANIMATION_FILENAME)

#     # We put the animation video file in the foreground of the original video
#     add_manim_animation_to_videoclip(original_video_filename, original_video_duration, TMP_MANIM_ANIMATION_FILENAME, output_filename)
    
# TODO: Is this below useful?
# export with transparent background: https://manimclass.com/manim-export-transparent-background/
# command to export:   manim --format=mp4 -qm -t Formula
